/*
 * Description: Enforce allocation offset tweaking to avoid CPU problems (4K-aliasing etc.)
 * Author: Catalin(ux) M. BOIE
 * E-mail: catab at embedromix dot ro
 * Web: http://kernel.embedromix.ro/us/
 */

#define __USE_GNU
#define	_GNU_SOURCE
#define __USE_XOPEN2K
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <dlfcn.h>

#include "cpublaster_config.h"

#define HASH_SIZE	32

extern __typeof (malloc)	__libc_malloc;
extern __typeof (realloc)	__libc_realloc;
extern __typeof (free)		__libc_free;

struct node
{
	void		*ptr;
	unsigned int	offset;
	struct node	*next;
};

struct info
{
	struct node	*head, *tail;
};


static unsigned int		offset;
static unsigned int		offset_increment = 64;
static unsigned int		offset_max = 4 * 4096;
static unsigned int		minim_allocation = 128 * 1024;
static unsigned int		verbose;
static struct info		hash[HASH_SIZE];
static FILE			*Log;


static struct node *get(const void *ptr)
{
	struct node *p;
	unsigned int h;

	if (ptr == NULL)
		return NULL;

	h = ((unsigned int ) ptr >> 16) % HASH_SIZE;

	p = hash[h].head;
	while (p != NULL) {
		if (p->ptr == ptr)
			return p;

		p = p->next;
	}

	return NULL;
}

static void add(void *ptr, unsigned int offset)
{
	struct node *p;
	unsigned int h;

	if (ptr == NULL)
		return;

	h = ((unsigned int) ptr >> 16) % HASH_SIZE;

	/* Try to find a free location */
	p = hash[h].head;
	while (p != NULL) {
		if (p->ptr == 0) {
			p->ptr = ptr;
			break;
		}

		p = p->next;
	}

	if (p == NULL) {
		p = (struct node *) __libc_malloc(sizeof(struct node));
		if (p == NULL)
			return;

		p->next = NULL;
		p->ptr = ptr;

		if (hash[h].tail == NULL) {
			hash[h].head = p;
		} else {
			hash[h].tail->next = p;
		}
		hash[h].tail = p;
	}
	p->offset = offset;
}

static int del(const void *ptr)
{
	struct node *p;
	unsigned int h;
	int ret = -1;

	if (ptr == NULL)
		return -1;

	h = ((unsigned int) ptr >> 16) % HASH_SIZE;

	p = hash[h].head;
	while (p != NULL) {
		if (p->ptr == ptr) {
			p->ptr = 0;
			ret = p->offset;
			break;
		}

		p = p->next;
	}

	return ret;
}

static void init(void)
{
	static unsigned char inited = 0;
	char *x;

	if (inited == 1)
		return;

	Log = stderr;

	x = getenv("CPUBLASTER_VERBOSE");
	if (x)
		verbose = strtol(x, NULL, 10);

	x = getenv("CPUBLASTER_MINIM_ALLOCATION");
	if (x)
		minim_allocation = strtol(x, NULL, 10);

	x = getenv("CPUBLASTER_INCREMENT");
	if (x)
		offset_increment = strtol(x, NULL, 10);

	x = getenv("CPUBLASTER_LOG");
	if (x) {
		Log = fopen(x, "a+");
		if (Log)
			setlinebuf(Log);
	}

	offset = (rand() % offset_max) / offset_increment * offset_increment;

	inited = 1;
}

void *malloc(size_t size)
{
	void *ret;

	init();

	if (size < minim_allocation)
		return __libc_malloc(size);

	ret = __libc_malloc(offset + size);
	if (ret == NULL)
		return NULL;

	if (verbose > 0)
		fprintf(Log, "malloc(%zu) offset=%u ret=%p\n",
			size, offset, ret + offset);

	ret += offset;

	add(ret, offset);

	offset += offset_increment;
	if (offset > offset_max)
		offset = offset_increment;

	return ret;
}

void *realloc(void *ptr, size_t size)
{
	struct node *p;
	void *new;

	p = get(ptr);
	if (p == NULL)
		return __libc_realloc(ptr, size);

	new = __libc_realloc(ptr - p->offset, p->offset + size);
	if (new == NULL)
		return NULL;

	new += p->offset;

	del(ptr);
	add(new, p->offset);

	if (verbose > 0)
		fprintf(Log, "realloc(%p, %zu) offset=%u ret=%p\n",
			ptr, size, p->offset, new);

	return new;
}

void free(void *ptr)
{
	int off;

	init();

	if (ptr == NULL)
		return;

	off = del(ptr);
	if (off != -1) {
		if (verbose > 0)
			fprintf(Log, "free(%p) off=%d new_ptr=%p\n",
				ptr, off, ptr - off);

		ptr -= off;
	}

	__libc_free(ptr);
}
